package nova.monadic_sfx.util

import scalafx.beans.property.ObjectProperty
import scalafx.beans.property.ReadOnlyObjectProperty
import scalafx.beans.value.ObservableValue

object Misc {

  implicit final class MyRichObservable[A, C](val self: ObservableValue[A, C])
      extends AnyVal {
    def filter(f: A => Boolean): ReadOnlyObjectProperty[A] =
      Method.filter(self)(f)
    def filterNull: ReadOnlyObjectProperty[A] = Method.filterNull(self)
  }

}
object Method {
  type Observable[A] = ObservableValue[A, _]

  def filter[B](
      a: Observable[B]
  )(f: B => Boolean): ReadOnlyObjectProperty[B] = {

    val prop = ObjectProperty[B](a.value)

    def changeHandler() =
      prop.synchronized {
        if (f(a.value)) {
          prop.value = a.value
        }
      }

    a onChange changeHandler()
    prop
  }

  /**
    * Simply creates a new observable that mirrors the source observable but
    * doesn't emit null values. JavaFX likes to work with null values in scene
    * nodes/properties (shrugs) and observables by default emit null values
    * that can cause crashes. ScalaFX does not offer any *fixes* for this
    *
    * @param a
    * @return
    */
  def filterNull[B](
      a: Observable[B]
  ): ReadOnlyObjectProperty[B] = {

    val prop = ObjectProperty[B](a.value)

    def changeHandler() =
      prop.synchronized {
        if (a.value != null) {
          prop.value = a.value
        }
      }

    a onChange changeHandler()
    prop
  }

}
